<?php

class RecordBookData {
    /** @var integer */
    public $id;
    /** @var string */
    public $externalID;
    /** @var integer */
    public $planID;
    /** @var string */
    public $planExternalID;
    /** @var integer */
    public $studentID;
    /** @var string */
    public $studentExternalID;
    /** @var integer */
    public $facultyID;
    /** @var string */
    public $facultyExternalID;
    /** @var string */
    public $status;
    /** @var string */
    public $degree;
    /** @var string */
    public $form;
    /** @var string */
    public $speciality;
    /** @var integer */
    public $grade;
    /** @var integer */
    public $group;
    /** @var integer */
    public $groupID;
}

class StudentData {
    /** @var integer */
    public $id;
    /** @var string */
    public $externalID;
    /** @var string */
    public $firstName;
    /** @var string */
    public $secondName;
    /** @var string */
    public $lastName;
    /** @var RecordBookData[] */
    public $recordBooks;
}

class Controller_Api_V0_Student extends Controller_Handler_Api {
    /** Нормализация данных зачетки
     * @param RecordBookData $recordBookData
     * @return RecordBookData
     */
    private function normalizeRecordBookData($recordBookData) {
        if (!isset($recordBookData->id) && isset($recordBookData->externalID)) {
            $recordBookData->id = Model_RecordBook::withExternalID($recordBookData->externalID);
        }

        // todo всегда false
        if (!isset($recordBookData->studentID)) {
            if (isset($recordBookData->studentExternalID)) {
                $recordBookData->studentID = Model_Student::withExternalID($recordBookData->studentExternalID);
                if ($recordBookData->studentID === -1) {
                    throw new InvalidArgumentException('RecordBook: student not found');
                }
            } else {
                throw new InvalidArgumentException('RecordBook: no student ids');
            }
        }

        if (!isset($recordBookData->planID)) {
            if (isset($recordBookData->planExternalID)) {
                $studyPlan = Model_Plan::withExternalID($recordBookData->planExternalID);
                if (is_null($studyPlan)) {
                    throw new InvalidArgumentException('RecordBook: plan not found');
                }
                $recordBookData->planID = $studyPlan->ID;
            } else {
                throw new InvalidArgumentException('RecordBook: no plan ids');
            }
        }

        if (!isset($recordBookData->facultyID)) {
            if (isset($recordBookData->facultyExternalID)) {
                $facultyID = $this->getFacultyIDByExternalID($recordBookData->facultyExternalID);
                if (is_null($facultyID)) {
                    throw new InvalidArgumentException('RecordBook: faculty not found');
                }
                $recordBookData->facultyID = $facultyID;
            } else {
                throw new InvalidArgumentException('RecordBook: no faculty ids');
            }
        }

        switch ($recordBookData->degree) {
            case 'Академический бакалавр':
            case 'Прикладной бакалавр':
            case 'Бакалавр':
                $recordBookData->degree = 'bachelor';
                break;

            case 'Специалист':
                $recordBookData->degree = 'specialist';
                break;

            case 'Магистр':
                $recordBookData->degree = 'master';
                break;

            case 'Аспирант':
//                $recordBookData->degree = 'postgraduate';
//                break;
                throw new InvalidArgumentException('Bad RecordBook degree');

            default:
                throw new InvalidArgumentException('Bad RecordBook degree');
        }

        switch ($recordBookData->status) {
            case 'Учится':
                $recordBookData->status = 'common';
                break;

            case 'Академ':
                $recordBookData->status = 'outlet';
                break;

            case 'Выпустился':
//                $recordBookData->status = 'expulsion';
//                break;
                throw new InvalidArgumentException('Bad RecordBook status');

            case 'Отчислен':
//                $recordBookData->status = 'leave';
//                break;
                throw new InvalidArgumentException('Bad RecordBook status');

            default:
                throw new InvalidArgumentException('Bad RecordBook status');
        }

        switch ($recordBookData->form) {
            case 'Очная':
                $recordBookData->form = '1';
                break;

            case 'Очно-заочная':
//                $recordBookData->form = '2';
//                break;
                throw new InvalidArgumentException('Bad RecordBook form');

            case 'Заочная':
//                $recordBookData->form = '3';
//                break;
                throw new InvalidArgumentException('Bad RecordBook form');

            default:
                throw new InvalidArgumentException('Bad RecordBook form');
        }

        if (!isset($recordBookData->groupID)) {
            if (isset($recordBookData->grade) && isset($recordBookData->degree)) {
                $gradeID = Model_Grades::getID($recordBookData->grade, $recordBookData->degree);
                if ($gradeID == -1) {
                    throw new InvalidArgumentException('RecordBook: grade not found');
                }

                $speciality = $recordBookData->speciality;
                if (strpos($speciality, ' - ') == 8) {
                    $speciality = substr($speciality, 11);
                }

                $year = Model_Plan::load($recordBookData->planID)->year;

                $group = Model_Group::findOrCreate($gradeID, $recordBookData->group, $speciality, $recordBookData->facultyID, $year);
                if (is_null($group)) {
                    throw new InvalidArgumentException('RecordBook: group not found');
                }

                $recordBookData->groupID = $group->getID();
            } else {
                throw new InvalidArgumentException('RecordBook has no grade or degree');
            }
        }

        return $recordBookData;
    }

    /** Нормализация данных студента
     * @param StudentData $studentData
     * @return StudentData
     */
    private function normalizeStudentData($studentData, $facultyID) {
        if (!isset($studentData->id) && isset($studentData->externalID)) {
            $studentData->id = Model_Student::withExternalID($studentData->externalID);
        }

        if (!isset($studentData->id) || ($studentData->id == -1)) {
            # todo: implement db procedure to get student by full name strictly
            $foundList = Model_Students::searchStudentsByName([$studentData->firstName,
                $studentData->secondName, $studentData->lastName], $facultyID);

            $foundSet = [];
            foreach ($foundList as $item) {
                if (trim($studentData->firstName) === trim($item['FirstName']) &&
                    trim($studentData->secondName) === trim($item['SecondName']) &&
                    trim($studentData->lastName) === trim($item['LastName'])) {
                    if (!in_array($item['ID'], $foundSet)) {
                        $foundSet[] = $item['ID'];
                    }
                }
            }

            if (count($foundSet) === 1) {
                $studentData->id = (int)$foundSet[0];
            } elseif (count($foundSet) > 1) {
                throw new InvalidArgumentException('Duplicate students');
            }
        }

        $del_flag = true;
        foreach ($studentData->recordBooks as &$recordBookData) {
            if (isset($studentData->id)) {
                $recordBookData->studentID = $studentData->id;
            }

            if (isset($studentData->externalID)) {
                $recordBookData->studentExternalID = $studentData->externalID;
            }

            try {
                $recordBookData = $this->normalizeRecordBookData($recordBookData);
            } catch (Exception $e) {
                Log::instance()->add(Log::WARNING, '{0} {1}: {2}', array(
                    '{0}' => 'SYNC_STUDENTS',
                    '{1}' => $e->getMessage(),
                    '{2}' => json_encode($recordBookData),
                ));
                $recordBookData = null;
            }

            if (isset($recordBookData))
                $del_flag = false;
        }

        if ($del_flag)
            $studentData = null;

        return $studentData;
    }

    /**
     * @param RecordBookData $recordBookData
     * @return integer
     */
    private function createOrUpdateRecordBook($recordBookData) {
        $recordBookBuilder = Model_RecordBook::make()
            ->studentID($recordBookData->studentID)
            ->planID($recordBookData->planID)
            ->group($recordBookData->groupID)
            ->externalID($recordBookData->externalID);

        // todo по коду встречаются "перевернутые" проверки, анализировать код сложно, необходимо сделать однородно
        if (isset($recordBookData->id) && $recordBookData->id > 0) {
            $recordBook = Model_RecordBook::load($recordBookData->id);
            $recordBook->update($recordBookBuilder->getPreparedData());
        } else {
            $recordBook = $recordBookBuilder->create();
        }

        return (int)$recordBook->ID;
    }

    /**
     * @param StudentData $studentData
     * @return integer[]
     */
    private function createOrUpdateStudent($studentData) {
        $existsBefore = false;
        // todo улучшиить нормализацию
        if (isset($studentData->id) && $studentData->id > 0) {
            $student = Model_Student::load($studentData->id);
            $existsBefore = true;

            // todo проверка отличается от остальных
            if (empty($student->ExternalID)) {
                $student->ExternalID = $studentData->externalID;
                $student->update();
            }
        } else {
            $student = Model_Student::make()
                ->externalID($studentData->externalID)
                ->firstName($studentData->firstName)
                ->secondName($studentData->secondName)
                ->lastName($studentData->lastName)
                ->create();
        }

        $recordBookIDs = [];

        foreach ($studentData->recordBooks as $recordBookData) {
            // todo перенести в нормализацию
            if (is_null($recordBookData)) continue;

            $recordBookData->studentID = $student->ID;

            try {
                // нужно поискать fake зачетные книжки
                if ($existsBefore) {
                    $oldRecordBooks = $student->getRecordBooks();
                    foreach ($oldRecordBooks as $rb) {
                        if (($recordBookData->degree == $rb['Degree']) &&
                            substr($rb['ExternalID'], 0, 5) == 'fake-') {
                            $recordBookData->id = (int)$rb['ID'];
                        }
                    }
                }
                $recordBookIDs[] = $this->createOrUpdateRecordBook($recordBookData);
            } catch (Exception $e) {
                Log::instance()->add(Log::WARNING, '{0} {1}: {2}', array(
                    '{0}' => 'SYNC_STUDENTS',
                    '{1}' => $e->getMessage(),
                    '{2}' => json_encode($recordBookData),
                ));
            }
        }

        return $recordBookIDs;
    }

//    public function action_get_index() {
//        $this->response->body('{ what: get index }');
//
//        return '{ what: get index }'; //$this->response;
//    }

//    /**
//     * @api {put} api/v0/student/recordBook Add new record book(s)
//     * @apiName Add new record book
//     * @apiGroup Record books
//     * @apiVersion 0.1.2
//     * @apiParam {String} token Api key
//     * @apiParam {String} [batch] Flag about massive addition
//     * @apiParam {String} externalID
//     * @apiParam {String} studentExternalID
//     * @apiParam {String} planExternalID
//     * @apiParam {String} facultyExternalID
//     * @apiParam {String} status
//     * @apiParam {String} level
//     * @apiParam {String} form
//     * @apiParam {String} speciality
//     * @apiParam {Int} grade
//     * @apiParam {Int} group
//     */
//    public function action_put_recordBook() {
//        try {
//            if ($this->request->query('batch') !== NULL) {
//                $data = json_decode($this->request->body());
//
//                $res = [];
//                foreach ($data as $item) {
//                    $data = $this->normalizeRecordBookData($item);
//                    $res[] = $this->createRecordBook($data);
//                }
//            } else {
//                /** @var RecordBookData $data */
//                $data = (object)$this->request->query();
//
//                $data = $this->normalizeRecordBookData($data);
//                $res[] = $this->createRecordBook($data);
//            }
//        } catch (Exception $e) {
//            $this->badRequestError($e->getMessage());
//        }
//
//        return $res;
//    }

//    /**
//     * @api {put} api/v0/student Add new student(s)
//     * @apiName Add new student
//     * @apiGroup Student
//     * @apiVersion 0.1.0
//     * @apiParam {String} token Api key
//     * @apiParam {String} [batch] Flag about massive addition
//     * @apiParam {String} firstName
//     * @apiParam {String} secondName
//     * @apiParam {String} lastName
//     * @apiParam {String} departmentName
//     * @apiParam {String} [departmentID]
//     * @apiParam {String} jobPositionName
//     * @apiParam {String} [jobPositionID]
//     * @apiParam {String} status
//     */
    public function action_put_index() {
        // TODO: проверять, что факультет специализации совпадает с факультетом учебного плана и facultyID
        $res = [];
        try {
            if ($this->request->query('batch') !== NULL) {
                $data = json_decode($this->request->body());
                $facultyID = Model_Faculties::getIdByExternalID($data->faculty);

                foreach ($data->students as $student) {
                    try {
                        $studentData = $this->normalizeStudentData($student, $facultyID);
                        if (isset($studentData))
                            $res[] = $this->createOrUpdateStudent($studentData);
                        else
                            $res[] = [];
                    } catch (Exception $e) {
                        Log::instance()->add(Log::WARNING, '{0} {1}: {2}', array(
                            '{0}' => 'SYNC_STUDENTS',
                            '{1}' => $e->getMessage(),
                            '{2}' => json_encode($student),
                        ));
                    }
                }

//                try {
//                    $student = $data->students[2];
//                    $res[] = $student;
//                    $studentData = $this->normalizeStudentData($student, $facultyID);
//                    $res[] = $this->createOrUpdateStudent($studentData);
//                } catch (Exception $e) {
//                    Log::instance()->add(Log::WARNING, '{0} {1}: {2}', array(
//                        '{0}' => 'SYNC_STUDENTS',
//                        '{1}' => $e->getMessage(),
//                        '{2}' => json_encode($student),
//                    ));
//                }
            } else {
//                $data = (object)$this->request->query();
//
//                $studentData = $this->normalizeStudentData($data);
//                $res[] = $this->createOrUpdateStudent($studentData);
                throw new InvalidArgumentException('No batch data received!');
            }
        } catch (Exception $e) {
            $this->badRequestError($e->getMessage());
        }

        return $res;
    }

//    public function action_get_list_by_name() {
//        return '{ what: nothing }';
//    }
}
